import snappy
import socket
import errno


class SnappySocket(object):
    def __init__(self, socket):
        self._decompressor = snappy.StreamDecompressor()
        self._compressor = snappy.StreamCompressor()
        self._socket = socket
        self._bootstrapped = None

    def __getattr__(self, name):
        return getattr(self._socket, name)

    def bootstrap(self, data):
        if data:
            self._bootstrapped = self._decompressor.decompress(data)

    def recv(self, size):
        return self._recv(size, self._socket.recv)

    def read(self, size):
        return self._recv(size, self._socket.read)

    def _recv(self, size, method):
        if self._bootstrapped:
            data = self._bootstrapped
            self._bootstrapped = None
            return data
        chunk = method(size)
        if chunk:
            uncompressed = self._decompressor.decompress(chunk)
        if not uncompressed:
            raise socket.error(errno.EWOULDBLOCK)
        return uncompressed

    def send(self, data):
        chunk = self._compressor.add_chunk(data, compress=True)
        self._socket.send(chunk)



import warnings
from reader import Reader


class LegacyReader(object):
    """
    In ``v0.5.0`` we dropped support for "tasks" in the :class:`nsq.Reader` API in
    favor of a single message handler.

    ``LegacyReader`` is a backwards compatible API for clients interacting with ``v0.5.0+`` that
    want to continue to use "tasks".

    Usage::

        from nsq import LegacyReader as Reader
    """
    def __init__(self, *args, **kwargs):
        warnings.warn('LegacyReader is a deprecated wrapper and will be removed in a future' +
                      ' release.  Use (multiple) Reader(s) each with their own' +
                      ' message handler.', DeprecationWarning)

        old_params = {}

        old_kwargs = ('all_tasks', 'topic', 'channel'
                      'nsqd_tcp_addresses', 'lookupd_http_addresses'
                      'max_tries', 'max_in_flight', 'requeue_delay', 'lookupd_poll_interval',
                      'low_rdy_idle_timeout', 'heartbeat_interval', 'max_backoff_duration')

        if args:
            keys = old_kwargs[:len(args)]
            old_params = dict(zip(keys, args))

        old_params.update(kwargs)

        all_tasks = old_params['all_tasks']
        topic = old_params['topic']
        channel = old_params['channel']

        del old_params['all_tasks']
        del old_params['topic']
        del old_params['channel']

        assert isinstance(all_tasks, dict)
        for key, method in all_tasks.items():
            assert callable(method), 'key %s must have a callable value' % key

        self.readers = []
        for task, method in all_tasks.items():
            if len(all_tasks) > 1:
                task_channel = channel + '.' + task
            else:
                task_channel = channel

            r = Reader(topic=topic, channel=task_channel, message_handler=method, **old_params)
            self.readers.append(r)



import signal
import tornado.ioloop
import logging

from nsq import Error, unpack_response, decode_message, valid_topic_name, valid_channel_name
from nsq import identify, subscribe, ready, finish, touch, requeue, nop, pub, mpub
from nsq import FRAME_TYPE_RESPONSE, FRAME_TYPE_ERROR, FRAME_TYPE_MESSAGE
from message import Message
from backoff_timer import BackoffTimer
from sync import SyncConn
from async import AsyncConn
from reader import Reader
from legacy_reader import LegacyReader
from writer import Writer
from version import __version__  # NOQA


def _handle_term_signal(sig_num, frame):
    logging.info('TERM Signal handler called with signal %r' % sig_num)
    tornado.ioloop.IOLoop.instance().stop()


def run():
    """
    Starts any instantiated :class:`nsq.Reader` or :class:`nsq.Writer`
    """
    signal.signal(signal.SIGTERM, _handle_term_signal)
    tornado.ioloop.IOLoop.instance().start()


__author__ = "Matt Reiferson <snakes@gmail.com>"
__all__ = ["Reader", "Writer", "run", "BackoffTimer", "Message", "Error", "LegacyReader",
           "SyncConn", "AsyncConn", "unpack_response", "decode_message",
           "identify", "subscribe", "ready", "finish", "touch", "requeue", "nop", "pub", "mpub",
           "valid_topic_name", "valid_channel_name",
           "FRAME_TYPE_RESPONSE", "FRAME_TYPE_ERROR", "FRAME_TYPE_MESSAGE"]



import time
import logging

import tornado.ioloop


class Client(object):
    def __init__(self):
        tornado.ioloop.PeriodicCallback(self._check_last_recv_timestamps, 60 * 1000).start()

    def _on_connection_identify(self, conn, data, **kwargs):
        logging.info('[%s:%s] IDENTIFY sent %r' % (conn.id, self.name, data))

    def _on_connection_identify_response(self, conn, data, **kwargs):
        logging.info('[%s:%s] IDENTIFY received %r' % (conn.id, self.name, data))
        if self.tls_v1 and not data.get('tls_v1'):
            logging.warning('[%s:%s] tls_v1 requested but disabled, could not negotiate feature',
                            conn.id, self.name)
        if self.snappy and not data.get('snappy'):
            logging.warning('[%s:%s] snappy requested but disabled, could not negotiate feature',
                            conn.id, self.name)

    def _on_connection_error(self, conn, error, **kwargs):
        logging.error('[%s:%s] ERROR: %r', conn.id, self.name, error)

    def _check_last_recv_timestamps(self):
        now = time.time()

        def is_stale(conn):
            timestamp = conn.last_recv_timestamp
            return (now - timestamp) > ((self.heartbeat_interval * 2) / 1000.0)

        # first get the list of stale connections, then close
        # (`conn.close()` may modify the list of connections while we're iterating)
        stale_connections = [conn for conn in self.conns.values() if is_stale(conn)]
        for conn in stale_connections:
            timestamp = conn.last_recv_timestamp
            # this connection hasnt received data for more than
            # the configured heartbeat interval, close it
            logging.warning('[%s:%s] connection is stale (%.02fs), closing',
                            conn.id, self.name, (now - timestamp))
            conn.close()
